@@ -13,12 +13,12 @@ Copyright 2008 Heikki Toivonen. All rights reserved.
 import logging
 import socket
 
-from M2Crypto import BIO, X509, m2, util  # noqa
+from M2Crypto import BIO, Err, X509, m2, py27plus, six, util  # noqa
 from M2Crypto.SSL import Checker, Context, timeout  # noqa
 from M2Crypto.SSL import SSLError
 from M2Crypto.SSL.Cipher import Cipher, Cipher_Stack
 from M2Crypto.SSL.Session import Session
-if util.py27plus:
+if py27plus:
     from typing import Any, AnyStr, Callable, Dict, List, Optional, Tuple, Union  # noqa
 
 __all__ = ['Connection',
@@ -29,12 +29,11 @@ log = logging.getLogger(__name__)
 
 
 def _serverPostConnectionCheck(*args, **kw):
-    # type: (*List[Any], **Dict[Any, Any]) -> int
+    # type: (*Any, **Any) -> int
     return 1
 
 
-class Connection:
-
+class Connection(object):
     """An SSL connection."""
 
     clientPostConnectionCheck = Checker.Checker()
@@ -42,17 +41,18 @@ class Connection:
 
     m2_bio_free = m2.bio_free
     m2_ssl_free = m2.ssl_free
+    m2_bio_noclose = m2.bio_noclose
 
     def __init__(self, ctx, sock=None, family=socket.AF_INET):
         # type: (Context, socket.socket, int) -> None
         """
 
-        @param ctx: SSL.Context
-        @param sock: socket to be used
-        @param family: socket family
+        :param ctx: SSL.Context
+        :param sock: socket to be used
+        :param family: socket family
         """
         self.ctx = ctx
-        self.ssl = m2.ssl_new(self.ctx.ctx)
+        self.ssl = m2.ssl_new(self.ctx.ctx)  # type: bytes
         if sock is not None:
             self.socket = sock
         else:
@@ -70,13 +70,18 @@ class Connection:
             self.set_post_connection_check_callback(
                 self.ctx.post_connection_check)
 
+        self.host = None
+
     def __del__(self):
         # type: () -> None
+        # Notice that M2Crypto doesn't automatically shuts down the
+        # connection here. You have to call self.close() in your
+        # program, M2Crypto won't do it automatically for you.
         if getattr(self, 'sslbio', None):
             self.m2_bio_free(self.sslbio)
         if getattr(self, 'sockbio', None):
             self.m2_bio_free(self.sockbio)
-        if self.ssl_close_flag == m2.bio_noclose and \
+        if self.ssl_close_flag == self.m2_bio_noclose and \
                 getattr(self, 'ssl', None):
             self.m2_ssl_free(self.ssl)
         self.socket.close()
@@ -115,7 +120,7 @@ class Connection:
         SSL_SENT_SHUTDOWN and SSL_RECEIVED_SHUTDOWN can be set at the
         same time.
 
-        @param mode: set the mode bitmask.
+        :param mode: set the mode bitmask.
         """
         m2.ssl_set_shutdown1(self.ssl, mode)
 
@@ -151,8 +156,8 @@ class Connection:
         will be called (for both the reading and writing side, if
         different).
 
-        @param readbio: BIO for reading
-        @param writebio: BIO for writing.
+        :param readbio: BIO for reading
+        :param writebio: BIO for writing.
         """
         m2.ssl_set_bio(self.ssl, readbio._ptr(), writebio._ptr())
 
@@ -165,11 +170,13 @@ class Connection:
 
         Makes sense only for servers.
 
-        @param cafile: Filename from which to load the CA list.
-        @return: 0 A failure while manipulating the STACK_OF(X509_NAME)
+        :param cafile: Filename from which to load the CA list.
+
+        :return: 0 A failure while manipulating the STACK_OF(X509_NAME)
                    object occurred or the X509_NAME could not be
                    extracted from cacert. Check the error stack to find
                    out the reason.
+
                  1 The operation succeeded.
         """
         m2.ssl_set_client_CA_list_from_file(self.ssl, cafile)
@@ -195,7 +202,7 @@ class Connection:
         By default, SSL struct will be freed in __del__. Call with
         m2.bio_close to override this default.
 
-        @param flag: either m2.bio_close or m2.bio_noclose
+        :param flag: either m2.bio_close or m2.bio_noclose
         """
         if flag not in (m2.bio_close, m2.bio_noclose):
             raise ValueError("flag must be m2.bio_close or m2.bio_noclose")
@@ -230,7 +237,7 @@ class Connection:
         The communication channel must already have been set and
         assigned to the ssl by setting an underlying BIO.
 
-        @return: 0 The TLS/SSL handshake was not successful but was shut
+        :return: 0 The TLS/SSL handshake was not successful but was shut
                    down controlled and by the specifications of the
                    TLS/SSL protocol. Call get_error() with the return
                    value ret to find out the reason.
@@ -256,10 +263,11 @@ class Connection:
         connection object and addr is the address bound to the other end
         of the SSL connection.
 
-        @return: tuple of Connection and addr. Address can take very
-        various forms (see socket documentation), for IPv4 it is
-        tuple(str, int), for IPv6 a tuple of four (host, port, flowinfo,
-        scopeid), where the last two are optional ints.
+        :return: tuple of Connection and addr. Address can take very
+                 various forms (see socket documentation), for IPv4 it
+                 is tuple(str, int), for IPv6 a tuple of four (host,
+                 port, flowinfo, scopeid), where the last two are
+                 optional ints.
         """
         sock, addr = self.socket.accept()
         ssl = Connection(self.ctx, sock)
@@ -288,8 +296,9 @@ class Connection:
         # type: (util.AddrType) -> int
         """Overloading socket.connect()
 
-        @param addr: addresses have various depending on their type
-        @return:status of ssl_connect()
+        :param addr: addresses have various depending on their type
+
+        :return:status of ssl_connect()
         """
         self.socket.connect(addr)
         self.addr = addr
@@ -299,7 +308,8 @@ class Connection:
         check = getattr(self, 'postConnectionCheck',
                         self.clientPostConnectionCheck)
         if check is not None:
-            if not check(self.get_peer_cert(), self.addr[0]):
+            if not check(self.get_peer_cert(),
+                         self.host if self.host else self.addr[0]):
                 raise Checker.SSLVerificationError(
                     'post connection check failed')
         return ret
@@ -345,6 +355,46 @@ class Connection:
         return self._write_nbio(data)
     sendall = send = write
 
+    def _decref_socketios(self):
+        pass
+
+    def recv_into(self, buff, nbytes=0):
+        # type: (Union[bytearray, memoryview], int) -> int
+        """
+        A version of recv() that stores its data into a buffer rather
+        than creating a new string.  Receive up to buffersize bytes from
+        the socket.  If buffersize is not specified (or 0), receive up
+        to the size available in the given buffer.
+
+        If buff is bytearray, it will have after return length of the
+        actually returned number of bytes. If buff is memoryview, then
+        the size of buff won't change (it cannot), but all bytes after
+        the number of returned bytes will be NULL.
+
+        :param buffer: a buffer for the received bytes
+        :param nbytes: maximum number of bytes to read
+        :return: number of bytes read
+
+        See recv() for documentation about the flags.
+        """
+        n = len(buff) if nbytes == 0 else nbytes
+
+        if n <= 0:
+            raise ValueError('size <= 0')
+
+        # buff_bytes are actual bytes returned
+        buff_bytes = m2.ssl_read(self.ssl, n, self._timeout)
+        buflen = len(buff_bytes)
+
+        # memoryview type has been added in 2.7
+        if py27plus and isinstance(buff, memoryview):
+            buff[:buflen] = buff_bytes
+            buff[buflen:] = b'\x00' * (len(buff) - buflen)
+        else:
+            buff[:] = buff_bytes
+
+        return buflen
+
     def read(self, size=1024):
         # type: (int) -> bytes
         if self._timeout != 0.0:
@@ -365,7 +415,7 @@ class Connection:
         s.setblocking(0) is equivalent to s.settimeout(0.0);
         s.setblocking(1) is equivalent to s.settimeout(None).
 
-        @param mode: new mode to be set
+        :param mode: new mode to be set
         """
         self.socket.setblocking(mode)
         if mode:
@@ -389,7 +439,7 @@ class Connection:
         # type: (int, int, Optional[int]) -> Union[int, bytes]
         """Get the value of the given socket option.
 
-        @param level: level at which the option resides.
+        :param level: level at which the option resides.
                To manipulate options at the sockets API level, level is
                specified as socket.SOL_SOCKET. To manipulate options at
                any other level the protocol number of the appropriate
@@ -397,16 +447,19 @@ class Connection:
                to indicate that an option is to be interpreted by the
                TCP protocol, level should be set to the protocol number
                of socket.SOL_TCP; see getprotoent(3).
-        @param optname: The value of the given socket option is
+
+        :param optname: The value of the given socket option is
                described in the Unix man page getsockopt(2)). The needed
                symbolic constants (SO_* etc.) are defined in the socket
                module.
-        @param buflen: If it is absent, an integer option is assumed
+
+        :param buflen: If it is absent, an integer option is assumed
                and its integer value is returned by the function. If
                buflen is present, it specifies the maximum length of the
                buffer used to receive the option in, and this buffer is
                returned as a bytes object.
-        @return: Either integer or bytes value of the option. It is up
+
+        :return: Either integer or bytes value of the option. It is up
                  to the caller to decode the contents of the buffer (see
                  the optional built-in module struct for a way to decode
                  C structures encoded as byte strings).
@@ -417,20 +470,23 @@ class Connection:
         # type: (int, int, Union[int, bytes, None]) -> Optional[bytes]
         """Set the value of the given socket option.
 
-        @param level: same as with getsockopt() above
-        @param optname: same as with getsockopt() above
-        @param value: an integer or a string representing a buffer. In
+        :param level: same as with getsockopt() above
+
+        :param optname: same as with getsockopt() above
+
+        :param value: an integer or a string representing a buffer. In
                       the latter case it is up to the caller to ensure
                       that the string contains the proper bits (see the
                       optional built-in module struct for a way to
                       encode C structures as strings).
-        @return: None for success or the error handler for failure.
+
+        :return: None for success or the error handler for failure.
         """
         return self.socket.setsockopt(level, optname, value)
 
     def get_context(self):
-        # type: () -> SSL.Context
-        """Return the SSL.Context object associated with this connection."""
+        # type: () -> Context
+        """Return the Context object associated with this connection."""
         return m2.ssl_get_ssl_ctx(self.ssl)
 
     def get_state(self):
@@ -443,8 +499,8 @@ class Connection:
         established. It however can be of significant interest during
         the handshake.
 
-        @return: 6 letter string indicating the current state of the SSL
-        object ssl.
+        :return: 6 letter string indicating the current state of the SSL
+                 object ssl.
         """
         return m2.ssl_get_state(self.ssl)
 
@@ -484,9 +540,9 @@ class Connection:
         """Return the peer certificate chain; if the peer did not provide
         a certificate chain, return None.
 
-        @warning: The returned chain will be valid only for as long as the
-        connection object is alive. Once the connection object gets freed,
-        the chain will be freed as well.
+        :warning: The returned chain will be valid only for as long as the
+                  connection object is alive. Once the connection object
+                  gets freed, the chain will be freed as well.
         """
         c = m2.ssl_get_peer_cert_chain(self.ssl)
         if c is None:
@@ -495,7 +551,7 @@ class Connection:
         return X509.X509_Stack(c)
 
     def get_cipher(self):
-        # type: () -> Optional[SSL.Cipher]
+        # type: () -> Optional[Cipher]
         """Return an M2Crypto.SSL.Cipher object for this connection; if the
         connection has not been initialised with a cipher suite, return None.
         """
@@ -505,7 +561,7 @@ class Connection:
         return Cipher(c)
 
     def get_ciphers(self):
-        # type: () -> Optional[SSL:Cipher_Stack]
+        # type: () -> Optional[Cipher_Stack]
         """Return an M2Crypto.SSL.Cipher_Stack object for this
         connection; if the connection has not been initialised with
         cipher suites, return None.
@@ -518,7 +574,7 @@ class Connection:
     def get_cipher_list(self, idx=0):
         # type: (int) -> str
         """Return the cipher suites for this connection as a string object."""
-        return m2.ssl_get_cipher_list(self.ssl, idx)
+        return six.ensure_text(m2.ssl_get_cipher_list(self.ssl, idx))
 
     def set_cipher_list(self, cipher_list):
         # type: (str) -> int
@@ -527,7 +583,10 @@ class Connection:
 
     def makefile(self, mode='rb', bufsize=-1):
         # type: (AnyStr, int) -> socket._fileobject
-        return socket._fileobject(self, mode, bufsize)
+        if six.PY3:
+            return socket.SocketIO(self, mode)
+        else:
+            return socket._fileobject(self, mode, bufsize)
 
     def getsockname(self):
         # type: () -> util.AddrType
@@ -537,7 +596,7 @@ class Connection:
         for instance. (The format of the address returned depends
         on the address family -- see above.)
 
-        @return:socket's address as addr type
+        :return:socket's address as addr type
         """
         return self.socket.getsockname()
 
@@ -549,7 +608,7 @@ class Connection:
         for instance.
         On some systems this function is not supported.
 
-        @return:
+        :return:
         """
         return self.socket.getpeername()
 
@@ -557,15 +616,15 @@ class Connection:
         # type: (bytes) -> int
         ret = m2.ssl_set_session_id_context(self.ssl, id)
         if not ret:
-            raise SSLError(m2.err_reason_error_string(m2.err_get_error()))
+            raise SSLError(Err.get_error_message())
 
     def get_session(self):
-        # type: () -> SSL.Session
+        # type: () -> Session
         sess = m2.ssl_get_session(self.ssl)
         return Session(sess)
 
     def set_session(self, session):
-        # type: (SSL.Session) -> None
+        # type: (Session) -> None
         m2.ssl_set_session(self.ssl, session._ptr())
 
     def get_default_session_timeout(self):
@@ -573,33 +632,33 @@ class Connection:
         return m2.ssl_get_default_session_timeout(self.ssl)
 
     def get_socket_read_timeout(self):
-        # type: () -> SSL.timeout
+        # type: () -> timeout
         return timeout.struct_to_timeout(
             self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO,
                                    timeout.struct_size()))
 
     def get_socket_write_timeout(self):
-        # type: () -> SSL.timeout
+        # type: () -> timeout
         return timeout.struct_to_timeout(
             self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_SNDTIMEO,
                                    timeout.struct_size()))
 
     def set_socket_read_timeout(self, timeo):
-        # type: (SSL.timeout) -> None
+        # type: (timeout) -> None
         assert isinstance(timeo, timeout.timeout)
         self.socket.setsockopt(
             socket.SOL_SOCKET, socket.SO_RCVTIMEO, timeo.pack())
 
     def set_socket_write_timeout(self, timeo):
-        # type: (SSL.timeout) -> None
+        # type: (timeout) -> None
         assert isinstance(timeo, timeout.timeout)
         self.socket.setsockopt(
             socket.SOL_SOCKET, socket.SO_SNDTIMEO, timeo.pack())
 
     def get_version(self):
         # type: () -> str
-        "Return the TLS/SSL protocol version for this connection."
-        return m2.ssl_get_version(self.ssl)
+        """Return the TLS/SSL protocol version for this connection."""
+        return six.ensure_text(m2.ssl_get_version(self.ssl))
 
     def set_post_connection_check_callback(self, postConnectionCheck):  # noqa
         # type: (Callable) -> None
@@ -611,3 +670,8 @@ class Connection:
         extension.
         """
         m2.ssl_set_tlsext_host_name(self.ssl, name)
+
+    def set1_host(self, name):
+        # type: (bytes) -> None
+        """Set the requested hostname to check in the server certificate."""
+        self.host = name
